// Libraries:
#include <Wire.h>
#include <DS3231.h>   // https://github.com/NorthernWidget/DS3231 v1.0.2.  Compiles with warnings, has missing functions, etc. But it does the job.
#include <TimerOne.h> // https://playground.arduino.cc/Code/Timer1/ v1.1.0

// Lemoncascade
//  Use an otherwise fairly useless 8x8 Red LED matrix with two 74HC595's to show a countdown to the end of the year in seconds, base 9.
// Mark Wilson Dec 2020

// Change the orientation to be input pins DOWN (vs RIGHT)
#define INPUTS_DOWN

//               --- Schematic ---
//   Common Gnd & 5VDC connections omitted.
//
//   +-- Uno --+     +---- Matrix ----+
//   |         |     |                |
//   |      D10+-----+"D_out"=Data in!|
//   |       D9+-----+"CLK"  =Clock   |
//   |       D8+-----+"LTH"  =Latch!  |
//   |         |     |                |
//   |         |     +----------------+      +--- RTC ---+
//   |         |                             |           |
//   | (SCL) A5+-----------------------------+SCL        |
//   | (SDA) A4+-----------------------------+SDA        |
//   |         |                             |           |
//   +---------+                             +-----------+
//
//   Matrix is a "FUNDUINO" 8x8 Red LED matrix with two 74HC595's.
//   RTC is a DS3231

const int latchPin = 8;
const int clockPin = 9;
const int dataPin = 10;

#define PIN_PORT PORTB
#define LATCH_PORT_BIT 1
#define CLOCK_PORT_BIT 2
#define DATA_PORT_BIT  4

#define SET(  _reg, _bits) (_reg) |=  (_bits)
#define CLEAR(_reg, _bits) (_reg) &= ~(_bits)

// Bit data of LED matrix, the data being shown (visible) and the data being updated (working)
// Bit 0 of ledsA[0] is bottom left (when input is on right)
byte ledsA[8];
byte ledsB[8];
byte* pVisibleLEDs = ledsA;
byte* pWorkingLEDs = ledsB;
volatile byte ledsUpdateRow = 0;
volatile byte ledsUpdateCol = 0;

volatile byte secondsCounter = 0;
volatile long microSecondsCounter = 0;
const long isrPeriodMicroseconds = 100L;

unsigned long countdownSeconds = 0;
int displayedSecond = -1;

DateTime Now;
char buff[16];

unsigned long splash[5] = // MEW 2020
{
  0xFBA27777UL,
  0xAA2A1515UL,
  0xABAA7575UL,
  0xAA2A4545UL,
  0xABBE7777UL
};
  
int nonary[8];
int convertToNonary(unsigned long value)
{
  // Converts value to base 9 in nonary, returns the number of digts
  // Least significant digits in nonary[0]
  int digits = 0;
  do
  {
    nonary[digits++] = value % 9;
    value /= 9;
  } while (value);
  return digits;
}

void ledsISR()
{
  // Update the led display, we draw one at a time, quite fast
  CLEAR(PIN_PORT, LATCH_PORT_BIT | CLOCK_PORT_BIT);
  word value = 0;
  if (pVisibleLEDs[ledsUpdateRow] & (0x01 << ledsUpdateCol))
  {
    // The 16-bit value sent out goes to two 8-bit 74HC595's driving rows and columns.
    // The row/col labelling assumes input pins on the right and the matrix seated one way.
    // The low byte controls the row: 1 bit is set HIGH (+5VDC), the others are LOW (Gnd).
    // The high byte controls the col: 1 bit is set to LOW (Gnd), the others are HIGH (+5VDC).  
    // Only one LED at the intersection is lit (one at a time)
    // LED @ (0, 0) is bottom left, if the input pins are on the right.
    value = byte(~(0b10000000 >> ledsUpdateCol)) | word(0x0100 << ledsUpdateRow);
  }
  word mask = 1;
  // clock out 16 bits
  while (mask)
  {
    if (value & mask)
      SET(PIN_PORT, DATA_PORT_BIT);
    else
      CLEAR(PIN_PORT, DATA_PORT_BIT);
    SET(PIN_PORT, CLOCK_PORT_BIT);
    CLEAR(PIN_PORT, CLOCK_PORT_BIT);
    mask <<= 1;
  }
  // latch them into effect
  SET(PIN_PORT, LATCH_PORT_BIT);

  // advance to the next LED
  ledsUpdateCol++;
  if (ledsUpdateCol > 7)
  {
    ledsUpdateCol = 0;
    ledsUpdateRow++;
    if (ledsUpdateRow > 7)
      ledsUpdateRow = 0;
  }

  // keep track of the time
  microSecondsCounter += isrPeriodMicroseconds;
  if (microSecondsCounter >= 1000000L)
  {
    microSecondsCounter -= 1000000L;
    secondsCounter++;
  }
}

void computeCountdown()
{
  // https://www.timeanddate.com/counters/multicountdown.html
  DateTime NewYearsDay(Now.year() + 1, 1, 1);
  // DateTime::secondstime() is declared but not implemented!
  // DateTime::unixtime() will do since we only want a difference.
  countdownSeconds = NewYearsDay.unixtime() - Now.unixtime();
}

void transformRowCol(int& row, int& col)
{
  // Adjust the row & col for different orientations, if required
#ifdef INPUTS_DOWN  
  int temp = col;
  col = 7 - row;
  row = temp;
#endif  
}

void blankLEDs()
{
  // Send 0's to explicitly blank the display
  CLEAR(PIN_PORT, LATCH_PORT_BIT | CLOCK_PORT_BIT);
  word mask = 1;
  // clock out 16 0's
  CLEAR(PIN_PORT, DATA_PORT_BIT);
  while (mask)
  {
    SET(PIN_PORT, CLOCK_PORT_BIT);
    CLEAR(PIN_PORT, CLOCK_PORT_BIT);
    mask <<= 1;
  }
  SET(PIN_PORT, LATCH_PORT_BIT);
}

void updateLEDs()
{
  // Swap the matrix being displayed and the one being set
  byte* pTemp = pVisibleLEDs;
  cli();
  pVisibleLEDs = pWorkingLEDs;
  sei();
  pWorkingLEDs = pTemp;  
}

void setLED(int row, int col, bool on)
{
  // Set a single LED on or off
  transformRowCol(row, col);
  if (on)
    pWorkingLEDs[row] |= 0x01 << col;
  else
    pWorkingLEDs[row] &= ~(0x01 << col);
}

void clearLEDs(int value = 0)
{
  // Set all the LEDs
  memset(pWorkingLEDs, value, sizeof(ledsA));
}

byte getByte(char*& ptr)
{
  // Extract a 2-digit number from the pointed-to chars, advances ptr
  byte value = *ptr++ - '0';
  value = value*10 + (*ptr++ - '0');
  return value;
}

void setup() 
{
  pinMode(latchPin, OUTPUT);
  pinMode(clockPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  blankLEDs();  // as soon as we can
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);
  Serial.begin(38400);
  memset(ledsA, 0, sizeof(ledsA));
  memset(ledsB, 0, sizeof(ledsB));
  Wire.begin();
  
  Serial.println("Lemoncascade. Countdown to the New Year in seconds, base 9.");
  Serial.println("Typing '?' prints the date & time as YYMMDDhhmmss.");
  Serial.println("Typing \"=YYMMDDhhmmss\" sets the date & time.");
  Serial.print("Countdown seconds (decimal):");
  Serial.println(countdownSeconds);
  Serial.print("Countdown seconds (nonary):");
  int digits = convertToNonary(countdownSeconds);
  while (digits--)
  {
    Serial.print((char)(nonary[digits] + '0'));
  }
  Serial.println();
 
  Timer1.initialize();
  Timer1.attachInterrupt(ledsISR, isrPeriodMicroseconds);
  
  // Splash....
  // all on
  clearLEDs(0xFF);
  updateLEDs();
  delay(1000);
  // dissolve
  byte LFSR = 0x0A;
  for (int i = 0; i < 255; i++)
  {
    // taps @ 8 6 5 4, period 255
    byte bit  = ((LFSR >> 0) ^ (LFSR >> 2) ^ (LFSR >> 3) ^ (LFSR >> 4)) & 1;
    LFSR = (LFSR >> 1) | (bit << 7);
    if (LFSR < 64)
    {
      cli();
      pVisibleLEDs[LFSR / 8] &= ~(1 << (LFSR % 8)); // direct access!
      sei();
      delay(25);
    }
  }
  clearLEDs();
  updateLEDs();
  delay(500);
  // banner  
  for (int step = -8; step < 32 + 8; step++)
  {
    clearLEDs();
    for (int row = 0; row < 5; row++)
      for (int col = 0; col < 8; col++)
        setLED(7 - row - 1, col, ((col + step) >= 0)?splash[row] & (0x80000000UL >> (col + step)):false);
    updateLEDs();
    delay(250);
  }
  
  Timer1.detachInterrupt();
  Now = RTClib::now();
  computeCountdown();
  Timer1.attachInterrupt(ledsISR, isrPeriodMicroseconds);
}


void loop() 
{
  if (Serial.available()) 
  {
    // a command
    int command = Serial.read();
    if (command == '?')
    {
      // Display the time etc as "YYMMDDhhmmss"
      Timer1.detachInterrupt();
      blankLEDs();
      Now = RTClib::now();
      Timer1.attachInterrupt(ledsISR, isrPeriodMicroseconds);
      int year = Now.year() - 2000;
      ::sprintf(buff, "=%02i%02i%02i%02i%02i%02i", year, Now.month(), Now.day(), Now.hour(), Now.minute(), Now.second());
      Serial.println(buff);
    }
    else if (command == '=')
    {
      // Set the time etc as "YYMMDDhhmmss"
      Timer1.detachInterrupt(); // see re-synch below
      blankLEDs();
      delay(10);
      int idx = 0;
      while (Serial.available() && idx < int(sizeof(buff) - 1))
      {
        buff[idx++] = Serial.read();
        delay(10);
      }
      buff[idx] = 0;
      if (idx > 12)
      {
        DS3231 rtc;
        char* pStr = buff;
        rtc.setYear(getByte(pStr));
        rtc.setMonth(getByte(pStr)); 
        rtc.setDate(getByte(pStr)); 
        rtc.setHour(getByte(pStr)); 
        // (we omit the day of the week)
        rtc.setMinute(getByte(pStr)); 
        rtc.setSecond(getByte(pStr));
      }
      else
      {
        Serial.println("Bad input");
      }
      Timer1.attachInterrupt(ledsISR, isrPeriodMicroseconds);
    }
  }

  byte secondsNow = secondsCounter; 
  if (secondsNow != displayedSecond)
  {
    displayedSecond = secondsNow;
    countdownSeconds--;
    
    if (displayedSecond >= 30)
    {
      // Re-synch with clock every 30s, otherwise coast using the ISR to count the seconds
      // Seems timer1 messes up I2C (or vice versa), disable it briefly while we read the time
      Timer1.detachInterrupt();
      blankLEDs();  // LED off while we're not updating
      Now = RTClib::now();
      secondsCounter = 0;
      displayedSecond = -1;
      Timer1.attachInterrupt(ledsISR, isrPeriodMicroseconds);
      computeCountdown();
    }

    clearLEDs();
    if (countdownSeconds <= 43046720L) // = 88888888 (base 9)
    {
      // convert to nonary
      int digits = convertToNonary(countdownSeconds);
      // display as columns, most significant digit on the right
      for (int digit = 0; digit < digits; digit++)
      {
        for (int row = 0; row < nonary[digit]; row++)
        {
          setLED(row, digit + 8 - digits, true);
        }
      }
    }
    else
    {
      setLED(7, 0, true); // error
    }
    updateLEDs();
  }
}
